import functools
import logging

import msgpack
from qtpy.QtCore import QTimer, QObject

import bluesky_kafka

from ..qt.threading import create_worker

logger = logging.getLogger(name="bluesky_widgets.qt.kafka_dispatcher")

LOADING_LATENCY = 0.01


class QtRemoteDispatcher(QObject):
    """
    Dispatch documents received over the network from a Kafka broker.

    This is designed to be run in a Qt application.

    Parameters
    ----------
    topics: list
        List of topics as strings such as ["topic-1", "topic-2"]
    bootstrap_servers : str
        Comma-delimited list of Kafka server addresses as a string
        such as ``'127.0.0.1:9092'``
    group_id: str
        Required string identifier for Kafka Consumer group
    consumer_config: dict
        Override default configuration or specify additional configuration
        options to confluent_kafka.Consumer.
    polling_duration: float
        Time in seconds to wait for a message before running function
        work_while_waiting. Default is 0.05.
    deserializer: function, optional
        optional function to deserialize data. Default is msgpack.loads.
    parent_qobject: QObject
        optional parent in the QT sense

    Example
    -------
    Plot data from all documents generated by remote RunEngines.
    >>> d = RemoteDispatcher(
    >>>         topics=["abc.bluesky.documents", "ghi.bluesky.documents"],
    >>>         bootstrap_servers='localhost:9092',
    >>>         group_id="document-printers",
    >>>         consumer_config={
    >>>             "auto.offset.reset": "latest"
    >>>         }
    >>>    )
    >>> d.subscribe(stream_documents_into_runs(model.add_run))
    >>> d.start()  # launches periodic workers on background threads
    >>> d.stop()  # stop launching workers
    """

    def __init__(
        self,
        topics,
        bootstrap_servers,
        group_id,
        consumer_config=None,
        polling_duration=0.05,
        deserializer=msgpack.loads,
        parent_qobject=None,
    ):
        super().__init__(parent_qobject)
        self.closed = False
        self._timer = QTimer(self)
        self._dispatcher = bluesky_kafka.RemoteDispatcher(
            topics=topics,
            bootstrap_servers=bootstrap_servers,
            group_id=group_id,
            consumer_config=consumer_config,
            polling_duration=polling_duration,
            deserializer=deserializer,
        )
        self.subscribe = self._dispatcher.subscribe
        self._waiting_for_start = True
        self.worker = None

    def _receive_data(self, continue_polling=None):
        # TODO Think about back pressure.
        if continue_polling is None:

            def continue_polling_forever():
                return True

            continue_polling = continue_polling_forever

        while continue_polling():
            try:
                # there should maybe be a poll method on the dispatcher
                msg = self._dispatcher._bluesky_consumer.consumer.poll(
                    self._dispatcher._bluesky_consumer.polling_duration
                )
                if msg is None:
                    logger.debug("no message")
                    break
                elif msg.error():
                    logger.error("Kafka Consumer error: %s", msg.error())
                else:
                    try:
                        # there should be a more direct way to deserialize the message
                        (
                            name,
                            document,
                        ) = self._dispatcher._bluesky_consumer._deserializer(msg.value())
                        if self._waiting_for_start:
                            if name == "start":
                                self._waiting_for_start = False
                            else:
                                # We subscribed midstream and are seeing documents for
                                # which we do not have the full run. Wait for a 'start'
                                # doc.
                                logger.debug("keep waiting for a start document")
                                return
                        yield self._dispatcher._bluesky_consumer.consumer, msg.topic(), name, document
                    except Exception as exc:
                        logger.exception(exc)
            except Exception as exc:
                logger.exception(exc)

        logger.debug("polling loop has ended cleanly")

    def start(self, continue_polling=None):
        logger.debug("QtRemoteDispatcher.start")
        if self.closed:
            raise RuntimeError(
                "This RemoteDispatcher has already been "
                "started and interrupted. Create a fresh "
                "instance with {}".format(repr(self))
            )

        self._work_loop(continue_polling=continue_polling)

    def _work_loop(self, continue_polling=None):
        self.worker = create_worker(
            self._receive_data,
            continue_polling=continue_polling,
        )
        # Schedule this method to be run again after a brief wait.
        self.worker.finished.connect(
            lambda: self._timer.singleShot(
                int(LOADING_LATENCY),
                functools.partial(self._work_loop, continue_polling),
            )
        )
        self.worker.yielded.connect(self._process_result)

        self.worker.start()

    def _process_result(self, result):
        if result is None:
            return
        consumer, topic, name, document = result
        self._dispatcher.process_document(consumer=consumer, topic=topic, name=name, document=document)

    def stop(self):
        logger.debug("QtRemoteDispatcher.stop")
        self.closed = True



from warnings import warn

from ..plot_builders import Lines
from ..plot_specs import Figure, Axes
from .._heuristics import hinted_fields
from ._base import AutoPlotter


class AutoLines(AutoPlotter):
    """
    Construct figures with line plots automatically.

    The decision of which lines to plot is based on metadata and data shape.

    Examples
    --------

    >>> model = AutoLines()
    >>> from bluesky_widgets.jupyter.figures import JupyterFigures
    >>> view = JupyterFigures(model.figures)
    >>> model.add_run(run)
    """

    def __init__(self, *, max_runs=None):
        super().__init__()
        # Map (stream_name, x, tuple_of_tuple_of_ys) to line of Lines instances for each group of y.
        self._lines_instances = {}
        self._max_runs = max_runs
        self.figures.events.removed.connect(self._on_figure_removed)

    @property
    def max_runs(self):
        return self._max_runs

    @max_runs.setter
    def max_runs(self, value):
        if value is not None:
            for builders in self._lines_instances.values():
                for builder in builders:
                    builder.max_runs = value
        self._max_runs = value

    def _on_figure_removed(self, event):
        super()._on_figure_removed(event)
        figure = event.item
        for key, plot_builders in self._lines_instances.items():
            for plot_builder in plot_builders:
                if figure == plot_builder.figure:
                    key_to_remove = key
        self._lines_instances.pop(key_to_remove)

    def handle_new_stream(self, run, stream_name, **kwargs):
        """
        This is used internally and should not be called directly by user code.

        Given a run and stream name, add or update figures with line plots.

        Parameters
        ----------
        run : BlueskyRun
        stream_name : String
        """
        omit_single_point_plot = False
        cleanup_motor_heuristic = False
        start_doc = run.metadata["start"]
        plan_hints = start_doc.get("hints", {})

        # Prepare a guess about the dimensions (independent variables) in case
        # we need it.
        motors = start_doc.get("motors")
        if motors is not None:
            GUESS = [([motor], "primary") for motor in motors]
        else:
            GUESS = [(["time"], "primary")]

        # Ues the guess if there is not hint about dimensions.
        dimensions = plan_hints.get("dimensions")
        if dimensions is None:
            cleanup_motor_heuristic = True
            dimensions = GUESS

        # We can only cope with all the dimensions belonging to the same
        # stream unless we resample. We are not doing to handle that yet.
        if len(set(d[1] for d in dimensions)) != 1:
            cleanup_motor_heuristic = True
            dimensions = GUESS  # Fall back on our GUESS.
            warn("We are ignoring the dimensions hinted because we cannot combine streams.")

        # for each dimension, choose one field only
        # the plan can supply a list of fields. It's assumed the first
        # of the list is always the one plotted against
        dim_fields = [fields[0] for fields, stream_name in dimensions]

        # make distinction between flattened fields and plotted fields
        # motivation for this is that when plotting, we find dependent variable
        # by finding elements that are not independent variables
        all_dim_fields = [field for fields, stream_name in dimensions for field in fields]

        _, dim_stream = dimensions[0]

        # We only care about the first descriptor because we are not referencing
        # configuration.
        stream = run[stream_name]
        descriptor = stream._descriptors[0]  # HACK!

        columns = hinted_fields(descriptor)

        # ## This deals with old descriptoruments. ## #

        if stream_name == "primary" and cleanup_motor_heuristic:
            # We stashed object names in dim_fields, which we now need to
            # look up the actual fields for.
            cleanup_motor_heuristic = False
            fixed_dim_fields = []
            for obj_name in dim_fields:
                # Special case: 'time' can be a dim_field, but it's not an
                # object name. Just add it directly to the list of fields.
                if obj_name == "time":
                    fixed_dim_fields.append("time")
                    continue
                try:
                    fields = descriptor.get("hints", {}).get(obj_name, {})["fields"]
                except KeyError:
                    fields = descriptor["object_keys"][obj_name]
                fixed_dim_fields.extend(fields)
            dim_fields = fixed_dim_fields

        # Ensure that no independent variables ('dimensions') are
        # duplicated here.
        columns = [c for c in columns if c not in all_dim_fields]

        # ## DECIDE WHICH KIND OF PLOT CAN BE USED ## #

        if (start_doc.get("num_points") == 1) and (stream_name == dim_stream) and omit_single_point_plot:
            return []

        # This is a heuristic approach until we think of how to hint this in a
        # generalizable way.
        if stream_name == dim_stream:
            dim_fields = dim_fields
        else:
            dim_fields = ["time"]  # 'time' once LivePlot can do that

        ndims = len(dim_fields)
        if not 0 < ndims < 3:
            # we need 1 or 2 dims to do anything, do not make empty figures
            return []

        # if self._fig_factory:
        #     fig = self._fig_factory(fig_name)
        # else:
        #     fig = plt.figure(fig_name)

        # if not fig.axes:
        #     if len(columns) < 5:
        #         layout = (len(columns), 1)
        #     else:
        #         nrows = ncols = int(np.ceil(np.sqrt(len(columns))))
        #         while (nrows - 1) * ncols > len(columns):
        #             nrows -= 1
        #         layout = (nrows, ncols)
        #     if ndims == 1:
        #         share_kwargs = {'sharex': True}
        #     elif ndims == 2:
        #         share_kwargs = {'sharex': True, 'sharey': True}
        #     else:
        #         raise NotImplementedError("we now support 3D?!")

        #     fig_size = np.array(layout[::-1]) * 5
        #     fig.set_size_inches(*fig_size)
        #     fig.subplots(*map(int, layout), **share_kwargs)
        #     for ax in fig.axes[len(columns):]:
        #         ax.set_visible(False)

        # axes = fig.axes

        # ## LIVE PLOT AND PEAK ANALYSIS ## #

        if ndims == 1:
            (x_key,) = dim_fields
            key = (stream_name, x_key, (tuple(columns),))
            try:
                lines_instances = self._lines_instances[key]
            except KeyError:
                lines_instances = []
                axes_list = []
                for y_key in columns:
                    dtype = descriptor["data_keys"][y_key]["dtype"]
                    if dtype not in ("number", "integer"):
                        warn("Omitting {} from plot because dtype is {}" "".format(y_key, dtype))
                        continue
                    axes = Axes(x_label=x_key, title=y_key)
                    axes_list.append(axes)
                    lines_kwargs = {}
                    if self.max_runs is not None:
                        lines_kwargs["max_runs"] = self.max_runs
                    lines = Lines(
                        x=x_key,
                        ys=(y_key,),
                        needs_streams=(stream_name,),
                        axes=axes,
                        **lines_kwargs,
                    )
                    lines_instances.append(lines)
                if not axes_list:
                    return
                title = ", ".join((str(lines.ys[0]) for lines in lines_instances)) + f" vs. {x_key}"
                if len(title) > 15:
                    short_title = title[:15] + "..."
                else:
                    short_title = title
                figure = Figure(axes_list, title=title, short_title=short_title)
                self._lines_instances[key] = lines_instances
                self.plot_builders.extend(lines_instances)
                self.figures.append(figure)
            for lines in lines_instances:
                lines.add_run(run, **kwargs)

        elif ndims == 2:
            return []
            # # Decide whether to use LiveGrid or LiveScatter. LiveScatter is the
            # # safer one to use, so it is the fallback..
            # gridding = self._start_descriptor.get('hints', {}).get('gridding')
            # if gridding == 'rectilinear':
            #     self._live_grids[descriptor['uid']] = {}
            #     slow, fast = dim_fields
            #     try:
            #         extents = self._start_descriptor['extents']
            #         shape = self._start_descriptor['shape']
            #     except KeyError:
            #         warn("Need both 'shape' and 'extents' in plan metadata to "
            #                 "create LiveGrid.")
            #     else:
            #         data_range = np.array([float(np.diff(e)) for e in extents])
            #         y_step, x_step = data_range / [max(1, s - 1) for s in shape]
            #         adjusted_extent = [extents[1][0] - x_step / 2,
            #                             extents[1][1] + x_step / 2,
            #                             extents[0][0] - y_step / 2,
            #                             extents[0][1] + y_step / 2]
            #         for I_key, ax in zip(columns, axes):
            #             # MAGIC NUMBERS based on what tacaswell thinks looks OK
            #             data_aspect_ratio = np.abs(data_range[1]/data_range[0])
            #             MAR = 2
            #             if (1/MAR < data_aspect_ratio < MAR):
            #                 aspect = 'equal'
            #                 ax.set_aspect(aspect, adjustable='box')
            #             else:
            #                 aspect = 'auto'
            #                 ax.set_aspect(aspect, adjustable='datalim')

            #             live_grid = LiveGrid(shape, I_key,
            #                                     xlabel=fast, ylabel=slow,
            #                                     extent=adjusted_extent,
            #                                     aspect=aspect,
            #                                     ax=ax)

            #             live_grid('start', self._start_descriptor)
            #             live_grid('descriptor', descriptor)
            #             self._live_grids[descriptor['uid']][I_key] = live_grid
            # else:
            #     self._live_scatters[descriptor['uid']] = {}
            #     x_key, y_key = dim_fields
            #     for I_key, ax in zip(columns, axes):
            #         try:
            #             extents = self._start_descriptor['extents']
            #         except KeyError:
            #             xlim = ylim = None
            #         else:
            #             xlim, ylim = extents
            #         live_scatter = LiveScatter(x_key, y_key, I_key,
            #                                     xlim=xlim, ylim=ylim,
            #                                     # Let clim autoscale.
            #                                     ax=ax)
            #         live_scatter('start', self._start_descriptor)
            #         live_scatter('descriptor', descriptor)
            #         self._live_scatters[descriptor['uid']][I_key] = live_scatter

        else:
            raise NotImplementedError("we do not support 3D+ in BEC yet (and it should have bailed above)")



from collections import abc

from qtpy import QtCore
from qtpy.QtCore import (
    QAbstractItemModel,
    QModelIndex,
    Qt,
)
from qtpy.QtWidgets import QAbstractItemView, QTreeView

from databroker.core import BlueskyEventStream


class RunTree:
    """Lazily populate the tree as data is requested."""

    def __init__(self, bs_run):
        self.run = bs_run
        self.children = []

        uid = RunNode(self.run, "uid", self.run.metadata["start"]["uid"], None, self)
        start = RunNode(self.run, "start", "dict", self.run.metadata["start"], self)
        start.num_children = len(self.run.metadata["start"])
        stop = RunNode(self.run, "stop", "dict", self.run.metadata["stop"], self)
        stop.num_children = len(self.run.metadata["stop"])
        streams = RunNode(self.run, "streams", f"({len(self.run)})", None, self)
        streams.num_children = len(self.run)
        self.children = [uid, start, stop, streams]

    def count(self):
        return len(self.children)

    def child(self, row):
        if row >= 0 and row < len(self.children):
            return self.children[row]
        return None


class RunNode(object):
    def __init__(self, run, key, value, data=None, parent=None):
        self.run = run
        self.parent = parent
        self.key = key
        self.value = value
        self.data = data
        self.children = []
        self.num_children = 0

    def child_count(self):
        return self.num_children

    def child(self, row):
        """Lazily fill these in when we can..."""
        if len(self.children) == 0 and self.num_children != len(self.children):
            self.fill_children()

        if row >= 0 and row < len(self.children):
            return self.children[row]

    def child_number(self):
        return self.parent.children.index(self)

    def fill_streams(self):
        """Fill the streams in."""
        for stream in self.run:
            num_events = self.run.metadata["stop"]["num_events"]
            if stream in num_events:
                n = num_events[stream]
                if n == 0:
                    value = "0 events"
                if n == 1:
                    value = "1 event"
                else:
                    value = f"{str(n)} events"
            else:
                value = "null"
            child = RunNode(self.run, stream, value, self.run[stream], self)
            # Establish how many child nodes there will be.
            stream_keys = self.run[stream].metadata["descriptors"][0]["data_keys"]
            child.num_children = len(stream_keys) + 2

            self.children.append(child)

    def fill_stream(self, stream):
        # For now just display the keys in the stream.
        node = RunNode(self.run, "metadata", "dict", self.data.metadata, self)
        node.num_children = len(self.data.metadata) - 1
        self.children.append(node)
        descriptors = self.data.metadata["descriptors"]
        if len(descriptors) == 1:
            node = RunNode(self.run, "descriptors (1)", "", descriptors[0], self)
            node.num_children = len(descriptors[0])
        else:
            node = RunNode(self.run, f"descriptors ({len(descriptors)})", descriptors, self)
            node.num_children = len(descriptors)

        self.children.append(node)
        stream_keys = self.data.metadata["descriptors"][0]["data_keys"]
        for key in stream_keys:
            value = f"{stream_keys[key]['dtype']} {stream_keys[key]['shape']}"
            child = RunNode(self.run, key, value, None, self)
            self.children.append(child)

    def fill_dict(self, data):
        for key in self.data:
            if isinstance(self.data[key], abc.Mapping):
                value = ""
            elif key == "descriptors":
                # This is "lifted up" and so skipping so as not to repeat.
                continue
            elif isinstance(self.data[key], abc.Iterable):
                value = str(self.data[key])
            else:
                value = self.data[key]
            child = RunNode(self.run, key, value, self.data[key], self)
            if isinstance(self.data[key], abc.Mapping):
                child.num_children = len(self.data[key])
            self.children.append(child)

    def fill_children(self):
        """Handle special ones like streams first."""
        if self.key == "streams":
            self.fill_streams()
        elif self.data and isinstance(self.data, BlueskyEventStream):
            self.fill_stream(self.data)
        elif self.data and isinstance(self.data, abc.Mapping):
            self.fill_dict(self.data)


class TreeViewModel(QAbstractItemModel):
    """
    Qt model connecting our run model to Qt's model-view machinery
    """

    def __init__(self, bs_run, parent=None):
        super(TreeViewModel, self).__init__(parent)

        self._catalog = bs_run
        if bs_run is not None:
            self._run_tree = RunTree(bs_run)
        else:
            self._run_tree = None

    def setRun(self, bs_run):
        self.beginResetModel()
        self._catalog = bs_run
        if bs_run is not None:
            self._run_tree = RunTree(bs_run)
        else:
            self._run_tree = None
        self.endResetModel()

    def index(self, row, column, parent=QModelIndex()):
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():
            parentItem = self._run_tree
        else:
            parentItem = parent.internalPointer()

        child = parentItem.child(row)
        if child:
            return self.createIndex(row, column, child)
        else:
            print("index not created ", row, " -> ", child, parent.key)

        return QModelIndex()

    def parent(self, index):
        """Return the parent."""
        if not index.isValid:
            return QModelIndex()
        item = index.internalPointer()
        if not item:
            return QModelIndex()

        parent = item.parent
        if parent == self._run_tree:
            return QModelIndex()
        else:
            return self.createIndex(parent.child_number(), 0, parent)

        return QModelIndex()

    def rowCount(self, parent=QModelIndex()):
        if parent.isValid():
            return parent.internalPointer().child_count()
        elif not parent.isValid() and self._run_tree is not None:
            return self._run_tree.count()
        return 0

    def columnCount(self, parent=None):
        return 2

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role != Qt.DisplayRole:
            return super().headerData(section, orientation, role)
        if orientation == Qt.Horizontal and section < self.columnCount():
            if section == 0:
                return str("Key")
            else:
                return str("Value")
            return str("test")
        elif orientation == Qt.Vertical and section < self.rowCount():
            return section

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if index.isValid() and role == Qt.DisplayRole:
            if index.column() == 0:
                return index.internalPointer().key
            else:
                return index.internalPointer().value
        elif not index.isValid():
            return "root"
        return QtCore.QVariant()


class QtTreeView(QTreeView):
    """
    Tree view showing a run
    """

    def __init__(self, model, parent=None):
        super(QtTreeView, self).__init__(parent)
        self.model = model

        self.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.setSortingEnabled(False)
        self.setAlternatingRowColors(True)

        # Initialize model.
        self._abstract_item_model = TreeViewModel(model.run)
        self.setModel(self._abstract_item_model)

        # Listen for future changes to model.
        self.model.events.run.connect(self.on_run_changed)

    def on_run_changed(self, event):
        self._abstract_item_model = TreeViewModel(event.run)
        self.setModel(self._abstract_item_model)



"""
Run like:

python -m bluesky_widgets.examples.utils.stream_data_kafka

It will publish Bluesky documents to kafka running on localhost:9092
until stopped.
"""
import asyncio
import logging
from functools import partial

import msgpack
import msgpack_numpy as mpn

from bluesky import RunEngine
from bluesky_kafka import Publisher
from bluesky.plans import scan
from bluesky.plan_stubs import sleep
from ophyd.sim import motor, det

log = logging.getLogger(__name__)


def stream_example_data(quiet=False):

    bootstrap_servers = "127.0.0.1:9092"

    producer_config = {
        "acks": 1,
        "enable.idempotence": False,
        "request.timeout.ms": 1000,
    }

    kafka_publisher = Publisher(
        topic="widgets_test.bluesky.documents",
        key="widgets_test",
        bootstrap_servers=bootstrap_servers,
        producer_config=producer_config,
        serializer=partial(msgpack.dumps, default=mpn.encode),
    )

    motor.delay = 0.22
    det.kind = "hinted"

    def infinite_plan():
        while True:
            for i in range(1, 5):
                yield from sleep(2)
                yield from scan([det], motor, -1, 1, 5 * i)

    RE = RunEngine(loop=asyncio.new_event_loop())
    RE.subscribe(kafka_publisher)

    # Just as a convenience, avoid collission with scan_ids of runs in Catalog.
    RE.md["scan_id"] = 100

    try:
        RE(infinite_plan())
    finally:
        RE.halt()


if __name__ == "__main__":
    import sys

    quiet = (len(sys.argv) > 1) and (sys.argv[1] in ("--quiet", "-q"))

    handler = logging.StreamHandler()
    handler.setLevel("INFO")
    log.setLevel("DEBUG")
    log.addHandler(handler)
    stream_example_data(quiet)

    # clear kafka topic?



from contextlib import contextmanager

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QApplication

from .threading import wait_for_workers_to_quit


_our_app_name = None


def get_our_app_name():
    return _our_app_name


@contextmanager
def gui_qt(app_name):
    """Start a Qt event loop in which to run the application.

    Parameters
    ----------
    app_name: str

    Notes
    -----
    This context manager is not needed if running the app within an interactive
    IPython session. In this case, use the ``%gui qt`` magic command, or start
    IPython with the Qt GUI event loop enabled by default by using
    ``ipython --gui=qt``.
    """
    app = QApplication.instance()
    if not app:
        # automatically determine monitor DPI.
        # Note: this MUST be set before the QApplication is instantiated
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
        # if this is the first time the Qt app is being instantiated, we set
        # the name, so that we know whether to raise_ in Window.show()
        app = QApplication([app_name])
        global _our_app_name
        _our_app_name = app_name
    else:
        app._existed = True
    app.aboutToQuit.connect(wait_for_workers_to_quit)
    yield app
    # if the application already existed before this function was called,
    # there's no need to start it again.  By avoiding unnecessary calls to
    # ``app.exec_``, we avoid blocking.
    if app.applicationName() == app_name:
        app.exec_()



from datetime import datetime, timedelta

import pytest

from ..search import SearchInput, LOCAL_TIMEZONE


def test_instantiation():
    SearchInput()


def test_instantiation_with_fields():
    "Smoke test with fields defined."
    SearchInput(fields=["field_1", "field_2"])


def test_field_search():
    "Check field_search can only be updated and query is updated."
    s = SearchInput(fields=["field_1", "field_2"])

    events = []

    def cb(event):
        events.append(event)

    s.events.field_search_updated.connect(cb)
    with pytest.raises(TypeError):
        s.field_search["field_1"] = "a"
    s.field_search.update({"field_1": "a"})
    assert s.field_search == {"field_1": "a", "field_2": None}
    assert len(events) == 1
    assert "field_1" in s.query
    assert "field_2" not in s.query
    s.field_search.update({"field_2": "b"})
    assert s.field_search == {"field_1": "a", "field_2": "b"}
    assert len(events) == 2
    assert "field_1" in s.query
    assert "field_2" in s.query
    s.field_search.update({"field_1": ""})
    assert s.field_search == {"field_1": "", "field_2": "b"}
    assert len(events) == 3
    assert "field_1" not in s.query
    assert "field_2" in s.query
    s.field_search.update({"field_1": "a", "field_2": "bb"})
    assert s.field_search == {"field_1": "a", "field_2": "bb"}
    assert len(events) == 4
    assert "field_1" in s.query
    assert "field_2" in s.query


def test_since_datetime():
    s = SearchInput()

    events = []

    def cb(event):
        events.append(event)

    s.events.since.connect(cb)
    s.since = datetime(2015, 9, 5, tzinfo=LOCAL_TIMEZONE)
    assert s.since == datetime(2015, 9, 5, tzinfo=LOCAL_TIMEZONE)
    assert len(events) == 1
    assert "time" in s.query
    assert "$gte" in s.query["time"]


def test_since_timedelta():
    s = SearchInput()

    events = []

    def cb(event):
        events.append(event)

    s.events.since.connect(cb)
    s.since = timedelta(days=-5)
    assert s.since == timedelta(days=-5)
    assert len(events) == 1
    assert "time" in s.query
    assert "$gte" in s.query["time"]


def test_until_datetime():
    s = SearchInput()

    events = []

    def cb(event):
        events.append(event)

    s.events.until.connect(cb)
    s.until = datetime(2015, 9, 5, tzinfo=LOCAL_TIMEZONE)
    assert s.until == datetime(2015, 9, 5, tzinfo=LOCAL_TIMEZONE)
    assert len(events) == 1
    assert "time" in s.query
    assert "$lt" in s.query["time"]


def test_until_timedelta():
    s = SearchInput()

    events = []

    def cb(event):
        events.append(event)

    s.events.until.connect(cb)
    s.until = timedelta(days=-5)
    assert s.until == timedelta(days=-5)
    assert len(events) == 1
    assert "time" in s.query
    assert "$lt" in s.query["time"]


def test_clearing():
    s = SearchInput()

    s.until = (2015, 9, 5)
    s.until = None
    s.since = None
    assert "time" not in s.query


def test_time_validator():
    s = SearchInput()

    allowed = {timedelta(days=-1), timedelta(days=-7), timedelta(days=-30)}

    def time_validator(since=None, until=None):
        """
        Enforce that since and until are values that a UI can represent.

        This is an example similar to what will be used in the Qt UI.
        """
        now = timedelta()
        if isinstance(since, timedelta):
            if not (until is None or until == now):
                raise ValueError(
                    "This UI cannot express since=timedelta(...) unless until is timedelta() or None."
                )
            for item in allowed:
                if since == item:
                    break
            else:
                # No matches
                raise ValueError(
                    "This UI can only express since as a timedelta if it is "
                    f"one of {allowed}. The value {since} is not allowed"
                )

    s.time_validator = time_validator

    s.until = None
    s.since = timedelta(days=-1)
    s.since = timedelta(days=-7)
    s.since = timedelta(days=-30)
    with pytest.raises(ValueError):
        s.since = timedelta(days=-2)
    with pytest.raises(ValueError):
        s.until = datetime(2015, 9, 5, tzinfo=LOCAL_TIMEZONE)
    with pytest.raises(ValueError):
        s.until = timedelta(days=-1)



import matplotlib.pyplot as plt


from ...models.plot_specs import Figure, Axes
from ..figures import HeadlessFigure


def test_closing():
    axes = Axes()
    model = Figure(axes=(axes,), title="test")
    view = HeadlessFigure(model)
    assert plt.fignum_exists(view.figure.number)
    view.close_figure()
    assert not plt.fignum_exists(view.figure.number)



"""
Run like:

python -m bluesky_widgets.examples.headless_figures

and it will print to stdout the names of the figures that it creates, one per line
"""
import tempfile

from bluesky import RunEngine
from bluesky.plans import scan
from ophyd.sim import motor, det

from bluesky_widgets.utils.streaming import stream_documents_into_runs
from bluesky_widgets.models.auto_plot_builders import AutoLines
from bluesky_widgets.headless.figures import HeadlessFigures
from bluesky_widgets.examples.utils.generate_msgpack_data import get_catalog

model = AutoLines(max_runs=3)
view = HeadlessFigures(model.figures)

RE = RunEngine()
RE.subscribe(stream_documents_into_runs(model.add_run))


catalog = get_catalog()
scans = catalog.search({"plan_name": "scan"})
model.add_run(scans[-1], pinned=True)


def plan():
    for i in range(1, 5):
        yield from scan([det], motor, -1, 1, 1 + 2 * i)


RE(plan())


directory = tempfile.mkdtemp()
filenames = view.export_all(directory)
print("\n".join(f'"{filename}"' for filename in filenames))



import collections
import copy
import os.path
import pprint
import time
import importlib

from bluesky_live.event import EmitterGroup, Event
from bluesky_queueserver import bind_plan_arguments
from bluesky_queueserver.manager.conversions import spreadsheet_to_plan_list
from bluesky_queueserver_api.zmq import REManagerAPI as REManagerAPI_ZMQ
from bluesky_queueserver_api.http import REManagerAPI as REManagerAPI_HTTP

from bluesky_queueserver_api._defaults import default_user_group


class RunEngineClient:
    """
    Parameters
    ----------
    zmq_control_addr: str or None
        Address of ZMQ server (Run Engine Manager). If None, then the default address defined
        in RE Manager code is used. (Default address is ``tcp://localhost:60615``).
    zmq_info_addr: str or None
        ZMQ address of the socket used by RE Manager to publishe console output.
    http_server_uri: str or None
        URI of HTTP server, e.g. ``http://localhost:60610``. If the URI is passed, then HTTP server
        is used for communication with Queue Server. Otherwise the specified or default 0MQ control
        and info addresses are used.
    user_name : str
        Name of the user submitting the plan. The name is saved as a parameter of the queue item
        and identifies the user submitting the plan (may be important in multiuser systems).
    user_group: str
        Name of the user group. User group is saved as a parameter of a queue item. Each user group
        can be assigned permissions to use a restricted set of plans and pass a restricted set of
        devices as plan parameters. Groups and group permissions are defined in the file
        ``user_group_permissions.yaml`` (see documentation for RE Manager).
    """

    def __init__(
        self,
        zmq_control_addr=None,
        zmq_info_addr=None,
        http_server_uri=None,
        http_server_api_key=None,
        user_name="GUI Client",
        user_group=default_user_group,
    ):
        if http_server_uri and (zmq_control_addr or zmq_info_addr):
            raise ValueError(
                "Invalid parameters: mutually excluseve HTTP Server URI and 0MQ control and/or info "
                "address are specified simultaneously."
            )
        if http_server_uri:
            print(f"HTTP server address: {http_server_uri}")
            self._client = REManagerAPI_HTTP(http_server_uri=http_server_uri)
            if http_server_api_key:
                self._client.set_authorization_key(api_key=http_server_api_key)
        else:
            print(f"0MQ control server address: {zmq_control_addr or 'default'}")
            print(f"0MQ info server address: {zmq_info_addr or 'default'}")
            self._client = REManagerAPI_ZMQ(zmq_control_addr=zmq_control_addr, zmq_info_addr=zmq_info_addr)

        self.set_map_param_labels_to_keys()

        # Address of remote 0MQ socket used to publish RE Manager console output
        self._zmq_info_addr = zmq_info_addr
        self._stop_console_monitor = False

        # User name and group are hard coded for now
        self._user_name = user_name
        self._user_group = user_group

        self._re_manager_status = {}
        self._re_manager_connected = None
        self._re_manager_status_time = time.time()
        # Minimum period of status update (avoid excessive call frequency)
        self._re_manager_status_update_period = 0.2

        self._allowed_devices = {}
        self._allowed_devices_uid = ""
        self._allowed_plans = {}
        self._allowed_plans_uid = ""
        self._plan_queue_items = []
        # Dictionary key: item uid, value: item pos in queue:
        self._plan_queue_items_pos = {}
        self._running_item = {}
        self._plan_queue_uid = ""
        self._run_list = []
        self._run_list_uid = ""
        self._plan_history_items = []
        self._plan_history_uid = ""

        # List of UIDs of the selected queue items, [] if no items are selected
        self._selected_queue_item_uids = []
        # History items are addressed by position (there could be repeated UIDs in the history)
        #   Items in the history can not be moved or deleted, only added to the bottom, so
        #   using positions is consistent. Empty list - no items are selected
        self._selected_history_item_pos = []

        # Current directory (used for opening/saving files).
        self.current_dir = None
        # Set the variable to the name of the Queue Server Custom Module (if available)
        self.qserver_custom_module_name = None
        # List of spreadsheet data types
        self.plan_spreadsheet_data_types = None
        # Dictionary of additional parameters: key - parameter name, value - a dictionary with
        #   the following key/value pairs: "text" - text description of the parameter to use in the form,
        #   "values" -  a list or a tuple of values, "type" - type of the value.
        self.plan_spreadsheet_additional_parameters = {}

        # Indicates if operation of destroying an environment is activated.
        self._env_destroy_activated = False

        # TODO: in the future the list of allowed instructions should be requested from the server
        self._allowed_instructions = {
            "queue_stop": {
                "name": "queue_stop",
                "description": "Stop execution of the queue.",
            }
        }

        self.events = EmitterGroup(
            source=self,
            status_changed=Event,
            plan_queue_changed=Event,
            running_item_changed=Event,
            plan_history_changed=Event,
            allowed_devices_changed=Event,
            allowed_plans_changed=Event,
            queue_item_selection_changed=Event,
            history_item_selection_changed=Event,
            history_item_process=Event,
        )

    @property
    def re_manager_status(self):
        return self._re_manager_status

    @property
    def re_manager_connected(self):
        return self._re_manager_connected

    def clear(self):
        # Clear the queue.
        try:
            self._client.queue_clear()
        except Exception as ex:
            raise RuntimeError(f"Failed to clear the plan queue: {ex}") from ex

    def clear_connection_status(self):
        """
        This function is not expected to clear 'status', only 'self._re_manager_connected'.
        """
        self._re_manager_connected = None
        self.events.status_changed(
            status=self._re_manager_status,
            is_connected=self._re_manager_connected,
        )

    def manager_connecting_ops(self):
        """
        Sequence of additional operations that should be performed while connecting to RE Manager.
        """
        self.load_allowed_devices()
        self.load_allowed_plans()
        self.load_plan_queue()
        self.load_plan_history()

    def load_re_manager_status(self, *, unbuffered=False):
        # TODO: the new API include efficient implementation of status management, there is no
        #       need to manage status in the application. The following code should be rewritten
        #       to take advantage of the existing API features.
        if unbuffered or (time.time() - self._re_manager_status_time > self._re_manager_status_update_period):
            status = self._re_manager_status.copy()
            accessible = self._re_manager_connected
            try:
                new_manager_status = self._client.status()
                self._re_manager_status.clear()
                self._re_manager_status.update(new_manager_status)
                self._re_manager_connected = True

                new_queue_uid = self._re_manager_status.get("plan_queue_uid", "")
                if new_queue_uid != self._plan_queue_uid:
                    self.load_plan_queue()
                new_run_list_uid = self._re_manager_status.get("run_list_uid", "")
                if new_run_list_uid != self._run_list_uid:
                    self.load_run_list()
                new_history_uid = self._re_manager_status.get("plan_history_uid", "")
                if new_history_uid != self._plan_history_uid:
                    self.load_plan_history()
                new_allowed_plans_uid = self._re_manager_status.get("plans_allowed_uid", "")
                if new_allowed_plans_uid != self._allowed_plans_uid:
                    self.load_allowed_plans()
                new_allowed_devices_uid = self._re_manager_status.get("devices_allowed_uid", "")
                if new_allowed_devices_uid != self._allowed_devices_uid:
                    self.load_allowed_devices()

            except (self._client.RequestTimeoutError, self._client.RequestError, self._client.ClientError):
                self._re_manager_connected = False
            if (status != self._re_manager_status) or (accessible != self._re_manager_connected):
                # Status changed. Initiate the updates
                self.events.status_changed(
                    status=self._re_manager_status,
                    is_connected=self._re_manager_connected,
                )

    def load_allowed_devices(self):
        try:
            result = self._client.devices_allowed(user_group=self._user_group)
            if result["success"] is False:
                raise RuntimeError(f"Failed to load list of allowed devices: {result['msg']}")
            self._allowed_devices.clear()
            self._allowed_devices.update(result["devices_allowed"])
            self._allowed_devices_uid = result["devices_allowed_uid"]
            self.events.allowed_devices_changed(allowed_devices=self._allowed_devices)
        except Exception as ex:
            print(f"Exception: {ex}")

    def load_allowed_plans(self):
        try:
            result = self._client.plans_allowed(user_group=self._user_group)
            if result["success"] is False:
                raise RuntimeError(f"Failed to load list of allowed plans: {result['msg']}")
            self._allowed_plans.clear()
            self._allowed_plans.update(result["plans_allowed"])
            self._allowed_plans_uid = result["plans_allowed_uid"]
            self.events.allowed_plans_changed(allowed_plans=self._allowed_plans)
        except Exception as ex:
            print(f"Exception: {ex}")

    def load_plan_queue(self):
        try:
            result = self._client.queue_get()
            if result["success"] is False:
                raise RuntimeError(f"Failed to load queue: {result['msg']}")
            self._plan_queue_items.clear()
            self._plan_queue_items.extend(result["items"])
            self._running_item.clear()
            self._running_item.update(result["running_item"])
            self._plan_queue_uid = result["plan_queue_uid"]

            # The dictionary that relates item uids and their positions in the queue.
            #   Used to speed up computations during queue operations.
            self._plan_queue_items_pos = {
                item["item_uid"]: n for n, item in enumerate(self._plan_queue_items) if "item_uid" in item
            }

            # Deselect queue items that are not in the queue or are not part of the contiguous
            #   selection. The selection will be cleared when the table is reloaded, so save
            #   it in local variable.
            selected_uids = self.selected_queue_item_uids
            pos, uids = -1, []
            for uid in selected_uids:
                p = self.queue_item_uid_to_pos(uid)
                if p >= 0:
                    if (pos < 0) or ((p >= 0) and (p == pos + 1)):
                        pos = p
                        uids.append(uid)
                    else:
                        break
            self.selected_queue_item_uids = uids

            # Update the representation of the queue
            self.events.plan_queue_changed(
                plan_queue_items=self._plan_queue_items,
                selected_item_uids=self.selected_queue_item_uids.copy(),
            )
            self.events.running_item_changed(
                running_item=self._running_item,
                run_list=self._run_list,
            )

        except Exception as ex:
            print(f"Exception: {ex}")

    def load_run_list(self):
        try:
            result = self._client.re_runs()
            if result["success"] is False:
                raise RuntimeError(f"Failed to load run_list: {result['msg']}")
            self._run_list.clear()
            self._run_list.extend(result["run_list"])
            self._run_list_uid = result["run_list_uid"]

            self.events.running_item_changed(
                running_item=self._running_item,
                run_list=self._run_list,
            )

        except Exception as ex:
            print(f"Exception: {ex}")

    def load_plan_history(self):
        try:
            result = self._client.history_get()
            if result["success"] is False:
                raise RuntimeError(f"Failed to load history: {result['msg']}")
            self._plan_history_items.clear()
            self._plan_history_items.extend(result["items"])
            self._plan_history_uid = result["plan_history_uid"]

            # Deselect queue history if it does not exist in the queue
            #   Selection will be cleared when the table is reloaded, so save it in local variable
            selected_item_pos = self.selected_history_item_pos
            if selected_item_pos and (selected_item_pos[-1] >= len(self._plan_history_items)):
                selected_item_pos = []
                self.selected_history_item_pos = selected_item_pos

            self.events.plan_history_changed(
                plan_history_items=self._plan_history_items.copy(),
                selected_item_pos=self.selected_history_item_pos,
            )

        except Exception as ex:
            print(f"Exception: {ex}")

    # ============================================================================
    #                       Useful functions
    def get_allowed_plan_parameters(self, *, name):
        """
        Returns the dictionary of parameters for the plan with name ``name`` from
        the list of allowed plans. Returns ``None`` if plan is not found in the list

        Parameters
        ----------
        name : str
            name of the plan

        Returns
        -------
        dict or None
            dictionary of plan parameters or ``None`` if the plan is not in the list.
        """
        return self._allowed_plans.get(name, None)

    def get_allowed_instruction_parameters(self, *, name):
        """
        Returns the dictionary of parameters for the instruction with name ``name`` from
        the list of allowed instructions. Returns ``None`` if plan is not found in the list

        Parameters
        ----------
        name : str
            name of the instruction

        Returns
        -------
        dict or None
            dictionary of instruction parameters or ``None`` if the plan is not in the list.
        """
        return self._allowed_instructions.get(name, None)

    def get_allowed_plan_names(self):
        return list(self._allowed_plans.keys()) if self._allowed_plans else []

    def get_allowed_instruction_names(self):
        return list(("queue_stop",))

    # ============================================================================
    #                         Item representation

    def set_map_param_labels_to_keys(self, *, map_dict=None):
        """
        Set mapping between labels and item dictionary keys. Map is a dictionary where
        keys are label names (e.g. names of the columns of a table) and dictionaries are
        tuples of keys that show the location of the parameter in item dictionary, e.g.
        ``{"STATUS": ("result", "exit_status")}``. In most practical cases this function
        should not be called at all.

        Parameters
        ----------
        map_dict : dict or None
            Map dictionary or None to use the default dictionary

        Returns
        -------
        None
        """
        if (map_dict is not None) and not isinstance(map_dict, collections.abc.Mapping):
            raise ValueError(
                f"Incorrect type ('{type(map_dict)}') of the parameter 'map'. 'None' or 'dict' is expected"
            )

        _default_map = {
            "": ("item_type",),
            "Name": ("name",),
            "Parameters": ("kwargs",),
            "USER": ("user",),
            "GROUP": ("user_group",),
            "STATUS": ("result", "exit_status"),
        }
        map_dict = map_dict if (map_dict is not None) else _default_map
        self._map_column_labels_to_keys = map_dict

    def get_bound_item_arguments(self, item):
        item_args = item.get("args", [])
        item_kwargs = item.get("kwargs", {})
        item_type = item.get("item_type", None)
        item_name = item.get("name", None)

        try:
            if item_type == "plan":
                plan_parameters = self._allowed_plans.get(item_name, None)
                if plan_parameters is None:
                    raise RuntimeError(f"Plan '{item_name}' is not in the list of allowed plans")
                bound_arguments = bind_plan_arguments(
                    plan_args=item_args,
                    plan_kwargs=item_kwargs,
                    plan_parameters=plan_parameters,
                )
                # If the arguments were bound successfully, then replace 'args' and 'kwargs'.
                item_args = []
                item_kwargs = bound_arguments.arguments
        except Exception:
            # print(
            #     f"Failed to bind arguments (item_type='{item_type}', "
            #     f"item_name='{item_name}'). Exception: {ex}"
            # )
            pass

        return item_args, item_kwargs

    def get_item_value_for_label(self, *, item, label, as_str=True):
        """
        Returns parameter value of the item for given label (e.g. table column name). Returns
        value represented as a string if `as_str=True`, otherwise returns value itself. Raises
        `KeyError` if the label or parameter is not found. It is not guaranteed that item
        dictionaries always contain all parameters, so exception does not indicate an error
        and should be processed by application.

        Parameters
        ----------
        item : dict
            Dictionary containing item parameters
        label : str
            Label (e.g. table column name)
        as_str : boolean
            ``True`` - return string representation of the value, otherwise return the value

        Returns
        -------
        str
            column value represented as a string

        Raises
        ------
        KeyError
            label or parameter is not found in the dictionary
        """
        try:
            key_seq = self._map_column_labels_to_keys[label]
        except KeyError:
            raise KeyError(f"Label '{label}' is not found in the map dictionary")

        # Follow the path in the dictionary. 'KeyError' exception is raised if a key does not exist
        try:
            value = item
            if (len(key_seq) == 1) and (key_seq[-1] in ("args", "kwargs")):
                # Special case: combine args and kwargs to be displayed in one column
                value = {
                    "args": value.get("args", []),
                    "kwargs": value.get("kwargs", {}),
                }
            else:
                for key in key_seq:
                    value = value[key]
        except KeyError:
            raise KeyError(f"Parameter with keys {key_seq} is not found in the item dictionary")

        if as_str:
            key = key_seq[-1]

            s = ""
            if key in ("args", "kwargs"):
                value["args"], value["kwargs"] = self.get_bound_item_arguments(item)

                s_args, s_kwargs = "", ""
                if value["args"] and isinstance(value["args"], collections.abc.Iterable):
                    s_args = ", ".join(f"{v}" for v in value["args"])
                if value["kwargs"] and isinstance(value["kwargs"], collections.abc.Mapping):
                    s_kwargs = ", ".join(f"{k}: {v}" for k, v in value["kwargs"].items())
                s = ", ".join([_ for _ in [s_args, s_kwargs] if _])

            elif key == "args":
                if value and isinstance(value, collections.abc.Iterable):
                    s = ", ".join(f"{v}" for v in value)

            elif key == "item_type":
                # Print capitalized first letter of the item type ('P' or 'I')
                s_tmp = str(value)
                if s_tmp:
                    s = s_tmp[0].upper()

            else:
                s = str(value)

        else:
            s = value

        return s

    # ============================================================================
    #                         Queue operations

    @property
    def selected_queue_item_uids(self):
        return self._selected_queue_item_uids

    @selected_queue_item_uids.setter
    def selected_queue_item_uids(self, item_uids):
        if self._selected_queue_item_uids != item_uids:
            self._selected_queue_item_uids = item_uids.copy()
            self.events.queue_item_selection_changed(selected_item_uids=item_uids)

    def queue_item_uid_to_pos(self, item_uid):
        # Returns -1 if item was not found
        return self._plan_queue_items_pos.get(item_uid, -1)

    def queue_item_pos_to_uid(self, n_item):
        try:
            item_uid = self._plan_queue_items[n_item]["item_uid"]
        except Exception:
            item_uid = ""
        return item_uid

    def queue_item_by_uid(self, item_uid):
        """
        Returns deep copy of the item based on item UID or None if the item was not found.

        Parameters
        ----------
        item_uid : str
            UID of an item. If ``item_uid=""`` then None will be returned

        Returns
        -------
        dict or None
            Dictionary of item parameters or ``None`` if the item was not found
        """
        if item_uid:
            sel_item_pos = self.queue_item_uid_to_pos(item_uid)
            if sel_item_pos >= 0:
                return copy.deepcopy(self._plan_queue_items[sel_item_pos])
        return None

    def _queue_items_move(self, *, sel_items, ref_item, position):
        """
        Move the batch of selected items above or below the reference item. The target item
        should not belong to the batch of items.

        Parameters
        ----------
        sel_items : list
            the list of selected item UIDs
        ref_item : str
            UID of the reference item
        position : str
            "before" - the items are moved above the target item, "after" - below the traget item
        """
        supported_positions = ("before", "after")
        if position not in supported_positions:
            raise ValueError(f"Unsupported position: {position}, supported values: {supported_positions}")

        if not sel_items or (ref_item in sel_items):
            return  # Nothing to do

        params = {"uids": sel_items}
        if position == "before":
            params.update({"before_uid": ref_item})
        else:
            params.update({"after_uid": ref_item})

        response = self._client.item_move_batch(**params)
        self.load_re_manager_status(unbuffered=True)

        if response["success"]:
            try:
                sel_items = [_["item_uid"] for _ in response["items"]]
            except Exception as ex:
                print(f"Item UID is not found in some items returned by the Queue Server: {ex}")

            self.selected_queue_item_uids = sel_items

        else:
            raise RuntimeError(f"Failed to move the item: {response['msg']}")

    def queue_items_move_up(self):
        """
        Move the selected batch of items up by one position
        """
        n_items = len(self._plan_queue_items)
        n_sel_items = len(self.selected_queue_item_uids)
        if not n_items or not n_sel_items or (n_items - n_sel_items < 1):
            return

        item_uid = self.selected_queue_item_uids[0]
        n_item = self.queue_item_uid_to_pos(item_uid)
        if item_uid and (n_item > 0):
            n_item_above = n_item - 1
            item_uid_above = self.queue_item_pos_to_uid(n_item_above)
            self._queue_items_move(
                sel_items=self._selected_queue_item_uids, ref_item=item_uid_above, position="before"
            )

    def queue_items_move_down(self):
        """
        Move the selected batch of items down by one position
        """
        n_items = len(self._plan_queue_items)
        n_sel_items = len(self.selected_queue_item_uids)
        if not n_items or not n_sel_items or (n_items - n_sel_items < 1):
            return

        item_uid = self.selected_queue_item_uids[-1]
        n_item = self.queue_item_uid_to_pos(item_uid)
        if item_uid and (0 <= n_item < n_items - 1):
            n_item_below = n_item + 1
            item_uid_below = self.queue_item_pos_to_uid(n_item_below)
            self._queue_items_move(
                sel_items=self._selected_queue_item_uids, ref_item=item_uid_below, position="after"
            )

    def queue_items_move_in_place_of(self, uid_ref_item):
        """
        Move the selected batch of items in the queue so that the first item of the batch assumes
        the position of the reference item. This is a generic 'move' operation. The reference item
        must not be included in the selected batch.
        """
        n_items = len(self._plan_queue_items)
        n_sel_items = len(self.selected_queue_item_uids)
        if not n_items or not n_sel_items or (n_items - n_sel_items < 1):
            return

        sel_item_uid_top = self.selected_queue_item_uids[0]
        sel_item_uid_bottom = self.selected_queue_item_uids[-1]
        n_item_top = self.queue_item_uid_to_pos(sel_item_uid_top)
        n_item_bottom = self.queue_item_uid_to_pos(sel_item_uid_bottom)
        n_item_to_replace = self.queue_item_uid_to_pos(uid_ref_item)

        if (n_item_to_replace < n_item_top) or (n_item_to_replace > n_item_bottom):
            position = "before" if (n_item_to_replace < n_item_top) else "after"
            self._queue_items_move(
                sel_items=self._selected_queue_item_uids, ref_item=uid_ref_item, position=position
            )

    def queue_items_move_to_top(self):
        """
        Move the selected batch of items to the top of the queue
        """
        if not self._plan_queue_items:
            return
        self.queue_items_move_in_place_of(self._plan_queue_items[0].get("item_uid", ""))

    def queue_items_move_to_bottom(self):
        """
        Move the selected batch of items to the bottom of the queue
        """
        if not self._plan_queue_items:
            return
        self.queue_items_move_in_place_of(self._plan_queue_items[-1].get("item_uid", ""))

    def queue_items_remove(self):
        """
        Delete the selected batch of items from queue.
        """
        sel_item_uids = self.selected_queue_item_uids.copy()
        if sel_item_uids:
            # Find and set UID of an item that will be selected once the current item is removed
            sel_item_uid_top = sel_item_uids[0]
            sel_item_uid_bottom = sel_item_uids[-1]
            n_item_top = self.queue_item_uid_to_pos(sel_item_uid_top)
            n_item_bottom = self.queue_item_uid_to_pos(sel_item_uid_bottom)

            n_items = len(self._plan_queue_items)

            if n_items <= 1:
                n_sel_item_new = -1
            elif n_item_bottom < n_items - 1:
                n_sel_item_new = n_item_bottom + 1
            else:
                n_sel_item_new = n_item_top - 1

            sel_item_new_uid = self.queue_item_pos_to_uid(n_sel_item_new)
            if sel_item_new_uid:
                self.selected_queue_item_uids = [sel_item_new_uid]
            else:
                self.selected_queue_item_uids = []

            try:
                response = self._client.item_remove_batch(uids=sel_item_uids)
            except Exception:
                print(f"Failed to delete item: {response['msg']}")
            finally:
                self.load_re_manager_status(unbuffered=True)

    def queue_clear(self):
        """
        Clear the plan queue
        """
        try:
            self._client.queue_clear()
        except Exception as ex:
            raise RuntimeError(f"Failed to clear the queue: {ex}") from ex
        finally:
            self.load_re_manager_status(unbuffered=True)

    def queue_mode_loop_enable(self, enable):
        """
        Enable or disable LOOP mode of the queue
        """
        try:
            self._client.queue_mode_set(loop=bool(enable))
        except Exception as ex:
            raise RuntimeError(f"Failed to change plan queue mode: {ex}") from ex
        finally:
            self.load_re_manager_status(unbuffered=True)

    def queue_item_copy_to_queue(self):
        """
        Copy currently selected item to queue. Item is supposed to be selected in the plan queue.
        """
        sel_item_uids = self._selected_queue_item_uids
        sel_items = []
        for uid in sel_item_uids:
            pos = self.queue_item_uid_to_pos(uid)
            if uid and (pos >= 0):
                sel_items.append(self._plan_queue_items[pos])
        if sel_items:
            self.queue_item_add_batch(items=sel_items)

    def queue_item_add(self, *, item, params=None):
        """
        Add item to queue. This function should be called by all widgets that add items to queue.
        The new item is inserted after the selected item or to the back of the queue in case
        no item is selected. Optional dictionary `params` may be used to override the default
        behavior. E.g. ``params={"pos": "front"}`` will add the item to the font of the queue.
        See the documentation for ``queue_item_add`` 0MQ API of Queue Server.
        The new item becomes the selected item.
        """
        if self._selected_queue_item_uids:
            # Insert after the last item in the selected batch
            sel_item_uid = self._selected_queue_item_uids[-1]
        else:
            # No selection: push to the back of the queue
            sel_item_uid = None

        queue_is_empty = not len(self._plan_queue_items)
        if not params:
            if queue_is_empty or not sel_item_uid:
                # Push button to the back of the queue
                params = {}
            else:
                params = {"after_uid": sel_item_uid}

        # We are submitting a plan as a new plan, so all unnecessary data will be stripped
        #   and new item UID will be assigned.
        request_params = {
            "item": item,
            "user": self._user_name,
            "user_group": self._user_group,
        }
        request_params.update(params)

        try:
            response = self._client.item_add(**request_params)
        except Exception as ex:
            raise RuntimeError(f"Failed to add item to the queue: {ex}") from ex
        finally:
            self.load_re_manager_status(unbuffered=True)

        try:
            # The 'item' and 'item_uid' should always be included in the returned item in case of success.
            sel_item_uid = response["item"]["item_uid"]
        except KeyError as ex:
            print(
                f"Item or item UID is not found in the server response {pprint.pformat(response)}. "
                f"Can not update item selection in the queue table. Exception: {ex}"
            )
        self.selected_queue_item_uids = [sel_item_uid]

    def queue_item_update(self, *, item):
        """
        Update the existing plan in the queue. This function should be called by all widgets
        that are used to modify (edit) the existing queue items. The items are distinguished by
        item UID, so item UID in the submitted ``item`` must match UID of the existing queue item
        that is replaced. The modified item becomes a selected item.
        """
        # We are submitting a plan as a new plan, so all unnecessary data will be stripped
        #   and new item UID will be assigned.
        request_params = {
            "item": item,
            "user": self._user_name,
            "user_group": self._user_group,
            "replace": True,  # Generates new UID
        }
        try:
            response = self._client.item_update(**request_params)
        except Exception as ex:
            raise RuntimeError(f"Failed to add item to the queue: {ex}") from ex
        finally:
            self.load_re_manager_status(unbuffered=True)

        try:
            # The 'item' and 'item_uid' should always be included in the returned item in case of success.
            sel_item_uid = response["item"]["item_uid"]
        except KeyError as ex:
            print(
                f"Item or item UID is not found in the server response {pprint.pformat(response)}. "
                f"Can not update item selection in the queue table. Exception: {ex}"
            )
        self.selected_queue_item_uids = [sel_item_uid]

    def queue_item_add_batch(self, *, items, params=None):
        """
        Add a batch of items to queue. This function should be called by all widgets that
        add a batch of items to queue. The new set of items added to the back of the queue.
        inserted after the selected item or to the back of the queue in case no item is selected.
        Optional dictionary `params` may be used to override the default behavior.
        E.g. ``params={"pos": "front"}`` will add the item to the font of the queue.
        See the documentation for ``queue_item_add_batch`` 0MQ API of Queue Server.
        The newly inserted items becomes selected.
        """
        # Do nothing if no items are to be inserted
        if not items:
            return

        sel_item_uids = self.selected_queue_item_uids.copy()

        if sel_item_uids:
            # Insert after the last item in the selected batch
            sel_item_uid = sel_item_uids[-1]
        else:
            # No selection: push to the back of the queue
            sel_item_uid = None

        queue_is_empty = not len(self._plan_queue_items)
        if not params:
            if queue_is_empty or not sel_item_uid:
                # Push button to the back of the queue
                params = {}
            else:
                params = {"after_uid": sel_item_uid}

        # We are submitting the plans in the batch as new plans, so all unnecessary
        #   data will be stripped and the new item UID will be assigned.
        request_params = {
            "items": items,
            "user": self._user_name,
            "user_group": self._user_group,
        }
        request_params.update(params)

        try:
            response = self._client.item_add_batch(**request_params)
        except Exception as ex:
            raise RuntimeError(f"Failed to add the batch of item to the queue: {ex}") from ex
        finally:
            self.load_re_manager_status(unbuffered=True)

        try:
            # The 'item' and 'item_uid' should always be included in the returned item in case of success.
            sel_item_uids = [_["item_uid"] for _ in response["items"]]
        except KeyError as ex:
            print(
                f"Item or item UID is not found in some of the items returned by the server "
                f"{pprint.pformat(response)}. Can not update item selection in the queue table. "
                f"Exception: {ex}"
            )
        self.selected_queue_item_uids = sel_item_uids

    def queue_upload_spreadsheet(self, *, file_path, data_type=None, **kwargs):
        # ``kwargs``` are passed to the custom spreadsheet processing function
        # TODO: significant part of this function is duplication of the code from
        #   ``bluesky_queueserver.server.server``. Implement reusable function as part of
        #   Queue Server API.

        file_path = os.path.expanduser(file_path)
        file_path = os.path.abspath(file_path)
        _, f_name = os.path.split(file_path)

        with open(file_path, "rb") as f:
            custom_code_module_name = self.qserver_custom_module_name

            custom_code_module = None
            if custom_code_module_name:
                try:
                    print(f"Importing custom module '{custom_code_module_name}' ...")
                    # logger.info("Importing custom module '%s' ...", custom_code_module_name)
                    custom_code_module = importlib.import_module(custom_code_module_name.replace("-", "_"))
                    print(f"Module '{custom_code_module_name}' was imported successfully.")
                    # logger.info("Module '%s' was imported successfully.", custom_code_module_name)
                except Exception as ex:
                    print(f"Failed to import custom instrument module '{custom_code_module_name}': {ex}")
                    # logger.error("Failed to import custom instrument module '%s':
                    # %s", custom_code_module_name, ex)

            item_list = []
            processed = False
            if custom_code_module and ("spreadsheet_to_plan_list" in custom_code_module.__dict__):
                print("Processing spreadsheet using function from external module ...")
                # logger.info("Processing spreadsheet using function from external module ...")
                # Try applying  the custom processing function. Some additional useful data is passed to
                #   the function. Unnecessary parameters can be ignored.
                item_list = custom_code_module.spreadsheet_to_plan_list(
                    spreadsheet_file=f, file_name=f_name, data_type=data_type, user=self._user_name, **kwargs
                )
                # The function is expected to return None if it rejects the file (based on 'data_type').
                #   Then try to apply the default processing function.
                processed = item_list is not None

            if not processed:
                # Apply default spreadsheet processing function.
                # logger.info("Processing spreadsheet using default function ...")
                item_list = spreadsheet_to_plan_list(
                    spreadsheet_file=f, file_name=f_name, data_type=data_type, user=self._user_name
                )

            if item_list is None:
                raise RuntimeError("Failed to process the spreadsheet: unsupported data type or format")

            # Since 'item_list' may be returned by user defined functions, verify the type of the list.
            if not isinstance(item_list, (tuple, list)):
                raise ValueError(
                    f"Spreadsheet processing function returned value of '{type(item_list)}' "
                    f"type instead of 'list' or 'tuple'"
                )

            # Ensure, that 'item_list' is sent as a list
            item_list = list(item_list)

            # Set item type for all items that don't have item type already set (item list may contain
            #   instructions, but it is responsibility of the user to set item types correctly.
            #   By default an item is considered a plan.
            for item in item_list:
                if "item_type" not in item:
                    item["item_type"] = "plan"

            # logger.debug("The following plans were created: %s", pprint.pformat(item_list))

            self.queue_item_add_batch(items=item_list)

    # ============================================================================
    #                         History operations

    @property
    def selected_history_item_pos(self):
        return self._selected_history_item_pos

    @selected_history_item_pos.setter
    def selected_history_item_pos(self, item_pos):
        """
        Sets the list of selected item in history

        Parameters
        ----------
        item_pos : iterable
            List or tuple of indices of the selected history items. Empty list/tuple
            if no items are selected.
        """
        item_pos = list(item_pos)
        if self._selected_history_item_pos != item_pos:
            self._selected_history_item_pos = item_pos
            self.events.history_item_selection_changed(selected_item_pos=item_pos)

    def history_item_add_to_queue(self):
        """Copy the selected plan from history to the end of the queue"""
        selected_item_pos = self.selected_history_item_pos
        if selected_item_pos:
            history_items = [self._plan_history_items[_] for _ in selected_item_pos]
            self.queue_item_add_batch(items=history_items)

    def history_item_send_to_processing(self):
        """
        Emits the event ``history_item_process`` sending the currently selected
        item as a parameter. The function should be called in response to some user
        action on the selected item (e.g. double clicking the item). The event
        can may be received by a widget that performs some processing of the item,
        e.g. loading from data broker and plotting the experimental data
        """
        selected_item_pos = self.selected_history_item_pos
        if selected_item_pos:
            # Copy data before sending it for processing by another model
            history_item = copy.deepcopy(self._plan_history_items[selected_item_pos[0]])
            self.events.history_item_process(item=history_item)

    def history_clear(self):
        """
        Clear history
        """
        try:
            self._client.history_clear()
        except Exception as ex:
            raise RuntimeError(f"Failed to clear the history: {ex}") from ex
        finally:
            self.load_re_manager_status(unbuffered=True)

    # ============================================================================
    #                     Operations with running item

    def running_item_add_to_queue(self):
        """Copy the selected plan from history to the end of the queue"""
        if self._running_item:
            running_item = self._running_item.copy()
            self.queue_item_add(item=running_item)

    # ============================================================================
    #                  Operations with RE Environment

    def environment_open(self, timeout=0):
        """
        Open RE Worker environment. Blocks until operation is complete or timeout expires.
        If ``timeout=0``, then the function blocks until operation is complete.

        Parameters
        ----------
        timeout : float
            maximum time for the operation. Exception is raised if timeout expires.
            If ``timeout=0``, the function blocks until operation is complete.

        Returns
        -------
        None
        """
        # Check if RE Worker environment already exists and RE manager is idle.
        self.load_re_manager_status()
        status = self._re_manager_status
        if status["manager_state"] != "idle":
            raise RuntimeError(f"RE Manager state must be 'idle': current state: {status['manager_state']}")
        if status["worker_environment_exists"]:
            raise RuntimeError("RE Worker environment already exists")

        # Initiate opening of RE Worker environment
        try:
            self._client.environment_open()
        except Exception as ex:
            raise RuntimeError(f"Failed to open RE Worker environment: {ex}") from ex

        # Wait for the environment to be created.
        if timeout:
            t_stop = time.time() + timeout
            while True:
                self.load_re_manager_status()
                status2 = self._re_manager_status
                if status2["worker_environment_exists"] and status2["manager_state"] == "idle":
                    break
                if timeout and (time.time() > t_stop):
                    raise RuntimeError("Failed to start RE Worker: timeout occurred")
                time.sleep(0.5)

        self.activate_env_destroy(False)

    def environment_close(self, timeout=0):
        """
        Close RE Worker environment. Blocks until operation is complete or timeout expires.
        If ``timeout=0``, then the function blocks until operation is complete.

        Parameters
        ----------
        timeout : float
            maximum time for the operation. Exception is raised if timeout expires.
            If ``timeout=0``, the function blocks until operation is complete.

        Returns
        -------
        None
        """
        # Check if RE Worker environment already exists and RE manager is idle.
        self.load_re_manager_status()
        status = self._re_manager_status
        if status["manager_state"] != "idle":
            raise RuntimeError(f"RE Manager state must be 'idle': current state: {status['manager_state']}")
        if not status["worker_environment_exists"]:
            raise RuntimeError("RE Worker environment does not exist")

        # Initiate opening of RE Worker environment
        try:
            self._client.environment_close()
        except Exception as ex:
            raise RuntimeError(f"Failed to close RE Worker environment: {ex}") from ex

        # Wait for the environment to be created.
        if timeout:
            t_stop = time.time() + timeout
        while True:
            self.load_re_manager_status()
            status2 = self._re_manager_status
            if not status2["worker_environment_exists"] and status2["manager_state"] == "idle":
                break
            if timeout and (time.time() > t_stop):
                raise RuntimeError("Failed to start RE Worker: timeout occurred")
            time.sleep(0.5)

        self.activate_env_destroy(False)

    @property
    def env_destroy_activated(self):
        return self._env_destroy_activated

    def activate_env_destroy(self, state):
        self._env_destroy_activated = bool(state)
        self.events.status_changed(
            status=self._re_manager_status,
            is_connected=self._re_manager_connected,
        )

    def environment_destroy(self, timeout=0):
        """
        Destroy (unresponsive) RE Worker environment. The function is intended for the cases when
        the environment is unresponsive and can not be stopped using ``environment_close``.
        Blocks until operation is complete or timeout expires. If ``timeout=0``, then the function
        blocks until operation is complete.

        Parameters
        ----------
        timeout : float
            maximum time for the operation. Exception is raised if timeout expires.
            If ``timeout=0``, the function blocks until operation is complete.

        Returns
        -------
        None
        """
        if not self._env_destroy_activated:
            raise RuntimeError("'Destroy Environment' operation is not activated and can not be executed")

        # Check if RE Worker environment already exists and RE manager is idle.
        self.load_re_manager_status()
        status = self._re_manager_status
        if not status["worker_environment_exists"]:
            raise RuntimeError("RE Worker environment does not exist")

        # Initiate opening of RE Worker environment
        try:
            self._client.environment_destroy()
        except Exception as ex:
            raise RuntimeError(f"Failed to destroy RE Worker environment: {ex}") from ex

        # Wait for the environment to be created.
        if timeout:
            t_stop = time.time() + timeout
        while True:
            self.load_re_manager_status()
            status2 = self._re_manager_status
            if not status2["worker_environment_exists"] and status2["manager_state"] == "idle":
                break
            if timeout and (time.time() > t_stop):
                raise RuntimeError("Failed to start RE Worker: timeout occurred")
            time.sleep(0.5)

        self.activate_env_destroy(False)

    # ============================================================================
    #                        Queue Control

    def queue_start(self):
        try:
            self._client.queue_start()
        except Exception as ex:
            raise RuntimeError(f"Failed to start the queue: {ex}") from ex

    def queue_stop(self):
        try:
            self._client.queue_stop()
        except Exception as ex:
            raise RuntimeError(f"Failed to request stopping the queue: {ex}") from ex

    def queue_stop_cancel(self):
        try:
            self._client.queue_stop_cancel()
        except Exception as ex:
            raise RuntimeError(f"Failed to cancel request to stop the queue: {ex}") from ex

    # ============================================================================
    #                        RE Control

    def _wait_for_completion(self, *, condition, msg="complete operation", timeout=0):
        if timeout:
            t_stop = time.time() + timeout

        while True:
            self.load_re_manager_status()
            status = self._re_manager_status
            if condition(status):
                break
            if timeout and (time.time() > t_stop):
                raise RuntimeError(f"Failed to {msg}: timeout occurred")
            time.sleep(0.5)

    def re_pause(self, timeout=0, *, option):
        """
        Pause execution of a plan.

        Parameters
        ----------
        timeout : float
            maximum time for the operation. Exception is raised if timeout expires.
            If ``timeout=0``, the function blocks until operation is complete.

        option : str
            "immediate" or "deferred"
        Returns
        -------
        None
        """

        # Initiate opening of RE Worker environment
        try:
            self._client.re_pause(option=option)
        except Exception as ex:
            raise RuntimeError(f"Failed to pause the running plan: {ex}") from ex

        # NOTE: it is unlikely that we need to wait for completion of pausing operation.
        #   1. The operation may take significant time to complete, meanwhile GUI will stay
        #      frozen.
        #   2. Requesting pause multiple times does not cause any issues.
        #   3. Waiting for completion prevents users from requesting immediate pause once
        #      the deferred pause is requested. This means that if the scan is stuck and never
        #      reaches the next setpoint, the users can not try immediate pause or do any other
        #      operation without restarting GUI.

        # def condition(status):
        #     return status["manager_state"] in ("idle", "paused")

        # self._wait_for_completion(condition=condition, msg="pause the running plan", timeout=timeout)

    def re_resume(self, timeout=0):
        """
        Pause execution of a plan.

        Parameters
        ----------
        timeout : float
            maximum time for the operation. Exception is raised if timeout expires.
            If ``timeout=0``, the function blocks until operation is complete.

        Returns
        -------
        None
        """

        # Initiate opening of RE Worker environment
        try:
            self._client.re_resume()
        except Exception as ex:
            raise RuntimeError(f"Failed to resume the running plan: {ex}") from ex

        def condition(status):
            return status["manager_state"] in ("idle", "executing_queue")

        self._wait_for_completion(condition=condition, msg="resume execution of the plan", timeout=timeout)

    def _re_continue_plan(self, *, action, timeout=0):

        if action not in ("stop", "abort", "halt"):
            raise RuntimeError(f"Unrecognized action '{action}'")

        method = f"re_{action}"
        try:
            getattr(self._client, method)()
        except Exception as ex:
            raise RuntimeError(f"Failed to {action} the running plan: {ex}") from ex

        def condition(status):
            return status["manager_state"] == "idle"

        self._wait_for_completion(condition=condition, msg=f"{action} the plan", timeout=timeout)

    def re_stop(self, timeout=0):
        self._re_continue_plan(action="stop", timeout=timeout)

    def re_abort(self, timeout=0):
        self._re_continue_plan(action="abort", timeout=timeout)

    def re_halt(self, timeout=0):
        self._re_continue_plan(action="halt", timeout=timeout)

    # ============================================================================
    #                        RE Manager console output

    def start_console_output_monitoring(self):
        self._stop_console_monitor = False
        self._client.console_monitor.enable()
        # self._rco = ReceiveConsoleOutput(zmq_subscribe_addr=self._zmq_info_addr, timeout=200)

    def stop_console_output_monitoring(self):
        self._stop_console_monitor = True

    # def console_monitoring_thread(self, *, callback):
    def console_monitoring_thread(self):

        while True:
            try:
                payload = self._client.console_monitor.next_msg(timeout=0.2)
                time, msg = payload.get("time", None), payload.get("msg", None)
                return time, msg

            except self._client.RequestTimeoutError:
                pass
            except Exception as ex:
                print(f"Exception occurred: {ex}")

            if self._stop_console_monitor:
                self._client.console_monitor.disable_wait()
                break



from bluesky_widgets.models.run_engine_client import RunEngineClient
from bluesky_widgets.qt import Window

from .widgets import QtViewer
from .settings import SETTINGS

from qtpy.QtWidgets import QAction


class ViewerModel:
    """
    This encapsulates on the models in the application.
    """

    def __init__(self):
        self.run_engine = RunEngineClient(
            zmq_control_addr=SETTINGS.zmq_re_manager_control_addr,
            zmq_info_addr=SETTINGS.zmq_re_manager_info_addr,
            http_server_uri=SETTINGS.http_server_uri,
            http_server_api_key=SETTINGS.http_server_api_key,
        )


class Viewer(ViewerModel):
    """
    This extends the model by attaching a Qt Window as its view.

    This object is meant to be exposed to the user in an interactive console.
    """

    def __init__(self, *, show=True, title="Demo App"):
        # TODO Where does title thread through?
        super().__init__()

        self._widget = QtViewer(self)
        self._window = Window(self._widget, show=show)

        menu_bar = self._window._qt_window.menuBar()
        menu_item_control = menu_bar.addMenu("Control Actions")
        self.action_activate_env_destroy = QAction("Activate 'Destroy Environment'", self._window._qt_window)
        self.action_activate_env_destroy.setCheckable(True)
        self._update_action_env_destroy_state()
        self.action_activate_env_destroy.triggered.connect(self._activate_env_destroy_triggered)
        menu_item_control.addAction(self.action_activate_env_destroy)

        self._widget.model.run_engine.events.status_changed.connect(self.on_update_widgets)

    def _update_action_env_destroy_state(self):
        env_destroy_activated = self._widget.model.run_engine.env_destroy_activated
        self.action_activate_env_destroy.setChecked(env_destroy_activated)

    def _activate_env_destroy_triggered(self):
        env_destroy_activated = self._widget.model.run_engine.env_destroy_activated
        self._widget.model.run_engine.activate_env_destroy(not env_destroy_activated)

    def on_update_widgets(self, event):
        self._update_action_env_destroy_state()

    @property
    def window(self):
        return self._window

    def show(self):
        """Resize, show, and raise the window."""
        self._window.show()

    def close(self):
        """Close the window."""
        self._window.close()



import argparse
import os

from bluesky_widgets.qt import gui_qt

from .viewer import Viewer
from .settings import SETTINGS


def main(argv=None):
    parser = argparse.ArgumentParser(description="BlueSky Queue Monitor")
    parser.add_argument(
        "--zmq-control-addr",
        default=None,
        help="Address of control socket of RE Manager, e.g. tcp://localhost:60615. "
        "If the address is passed as a CLI parameter, it overrides the address specified with "
        "QSERVER_ZMQ_CONTROL_ADDRESS environment variable. Default address is "
        "used if the parameter or the environment variable are not specified.",
    )
    parser.add_argument(
        "--zmq-control",
        default=None,
        help="The parameter is deprecated and will be removed. Use --zmq-control-addr instead.",
    )
    parser.add_argument(
        "--zmq-info-addr",
        default=None,
        help="Address of PUB-SUB socket of RE Manager, e.g. 'tcp://localhost:60625'. "
        "If the address is passed as a CLI parameter, it overrides the address specified with "
        "QSERVER_ZMQ_INFO_ADDRESS environment variable. Default address is "
        "used if the parameter or the environment variable are not specified.",
    )
    parser.add_argument(
        "--zmq-publish",
        default=None,
        help="The parameter is deprecated and will be removed. Use --zmq-info-addr instead.",
    )
    parser.add_argument(
        "--http-server-uri",
        default=None,
        help="Address of HTTP Server, e.g. 'http://localhost:60610'. Activates communication "
        "with Queue Server via HTTP server. If the address is passed as a CLI parameter, "
        "it overrides the address specified with QSERVER_HTTP_SERVER_URI environment variable. "
        "Use QSERVER_HTTP_SERVER_API_KEY environment variable to pass an API key for authorization.",
    )
    args = parser.parse_args(argv)

    # The priority is first to check if an address is passed as a parameter and then
    #   check if the environment variables are set. Otherwise use the default local addresses.
    #   (The default addresses are typically used in demos.)
    zmq_control_addr = args.zmq_control_addr
    zmq_control_addr = zmq_control_addr or args.zmq_control
    if args.zmq_control is not None:
        print("The parameter --zmq-control is deprecated and will be removed. Use --zmq-control-addr instead.")
    zmq_control_addr = zmq_control_addr or os.environ.get("QSERVER_ZMQ_CONTROL_ADDRESS", None)

    zmq_info_addr = args.zmq_info_addr
    if args.zmq_publish is not None:
        print("The parameter --zmq-publish is deprecated and will be removed. Use --zmq-info-addr instead.")
    zmq_info_addr = zmq_info_addr or args.zmq_publish
    zmq_info_addr = zmq_info_addr or os.environ.get("QSERVER_ZMQ_INFO_ADDRESS", None)
    if "QSERVER_ZMQ_PUBLISH_ADDRESS" in os.environ:
        print("WARNING: Environment variable QSERVER_ZMQ_PUBLISH_ADDRESS is deprecated and will be removed.")
        print("    Use QSERVER_ZMQ_INFO_ADDRESS environment variable instead.")
    zmq_info_addr = zmq_info_addr or os.environ.get("QSERVER_ZMQ_PUBLISH_ADDRESS", None)

    http_server_uri = args.http_server_uri
    http_server_uri = http_server_uri or os.environ.get("QSERVER_HTTP_SERVER_URI", None)

    http_server_api_key = os.environ.get("QSERVER_HTTP_SERVER_API_KEY", None)

    if http_server_uri:
        print("Initializing: communication with Queue Server via HTTP Server ...")
        SETTINGS.http_server_uri = http_server_uri
        SETTINGS.http_server_api_key = http_server_api_key
        SETTINGS.zmq_re_manager_control_addr = None
        SETTINGS.zmq_re_manager_info_addr = None
    else:
        print("Initializing: communication with Queue Server directly via 0MQ sockets ...")
        SETTINGS.http_server_uri = None
        SETTINGS.http_server_api_key = None
        SETTINGS.zmq_re_manager_control_addr = zmq_control_addr
        SETTINGS.zmq_re_manager_info_addr = zmq_info_addr

    with gui_qt("BlueSky Queue Monitor"):
        viewer = Viewer()  # noqa: 401


if __name__ == "__main__":
    main()



from bluesky_live.run_builder import build_simple_run
import numpy

from .. import AutoImages
from ....headless.figures import HeadlessFigures


def test_images():
    "Test AutoImages with a 2D array."
    run = build_simple_run({"ccd": numpy.random.random((11, 13))})
    model = AutoImages()
    view = HeadlessFigures(model.figures)
    assert not model.figures
    model.add_run(run)
    assert len(model.figures) == 1
    assert model.figures[0].axes[0].artists
    view.close()


def test_images_multiple_fields():
    "Test AutoImages with multiple fields with varied shapes."
    run = build_simple_run(
        {
            "ccd": numpy.random.random((11, 13)),
            "ccd2": numpy.random.random((17, 19, 23)),
        }
    )
    model = AutoImages()
    view = HeadlessFigures(model.figures)
    assert not model.figures
    model.add_run(run)
    assert len(model.figures) == 2
    assert model.figures[0].axes[0].artists
    assert model.figures[1].axes[0].artists
    view.close()



"""
An example integration QtFigures into an "existing" Qt application. Run like:

python -m bluesky_widgets.examples.qt_app_integration
"""
from qtpy.QtWidgets import (
    QApplication,
    QPushButton,
    QVBoxLayout,
    QLabel,
    QMainWindow,
    QWidget,
)
from bluesky_widgets.models.search import Search
from bluesky_widgets.qt.search import QtSearch
from bluesky_live.event import Event


# Extend the search widget with a single button. In your application, you might
# want multiple buttons that do different things.


class SearchAndOpen(Search):
    """
    Extend Search model with a signal for when a result is "opened".

    In your application, you might have multiple such signals associated with
    different buttons.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.events.add(open=Event)

    @property
    def selected_runs(self):
        # This property would be useful in general and should be added to the
        # Search itself in bluesky-widgets.
        return [self.results[uid] for uid in self.selected_uids]


class QtSearchListWithButton(QWidget):
    """
    A view for SearchAndOpen.

    Combines the QtSearches widget with a button.
    """

    def __init__(self, model: SearchAndOpen, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model
        layout = QVBoxLayout()
        self.setLayout(layout)
        layout.addWidget(QtSearch(model))

        # Add a button that does something with the currently-selected Runs
        # when you click it.
        self._open_button = QPushButton("Open")
        layout.addWidget(self._open_button)

        # Register a callback (slot) for the button qt click signal.
        self._open_button.clicked.connect(self._on_click_open_button)

    def _on_click_open_button(self):
        """
        Receive the Qt signal and emit a bluesky-widgets one.

        Include a list of BlueskyRuns corresponding to the current selection.
        """
        self.model.events.open(selected_runs=self.model.selected_runs)


# Customize what is shown in the search results here.

headings = (
    "Unique ID",
    "Transient Scan ID",
    "Plan Name",
    "Scanning",
    "Start Time",
    "Duration",
    "Exit Status",
)


def extract_results_row_from_run(run):
    """
    Given a BlueskyRun, format a row for the table of search results.
    """
    from datetime import datetime

    metadata = run.describe()["metadata"]
    start = metadata["start"]
    stop = metadata["stop"]
    start_time = datetime.fromtimestamp(start["time"])
    if stop is None:
        str_duration = "-"
    else:
        duration = datetime.fromtimestamp(stop["time"]) - start_time
        str_duration = str(duration)
        str_duration = str_duration[: str_duration.index(".")]
    return (
        start["uid"][:8],
        start.get("scan_id", "-"),
        start.get("plan_name", "-"),
        str(start.get("motors", "-")),
        start_time.strftime("%Y-%m-%d %H:%M:%S"),
        str_duration,
        "-" if stop is None else stop["exit_status"],
    )


columns = (headings, extract_results_row_from_run)


def main():
    # First, some boilerplate to make a super-minimal Qt application that we want
    # to add some bluesky-widgets components into.
    app = QApplication(["Some App"])
    window = QMainWindow()
    central_widget = QWidget(window)
    window.setCentralWidget(central_widget)
    central_widget.setLayout(QVBoxLayout())
    central_widget.layout().addWidget(QLabel("This is part of the 'original' app."))
    window.show()

    # *** INTEGRATION WITH BLUESKY-WIDGETS STARTS HERE. ***

    # Ensure that any background workers started by bluesky-widgets stop
    # gracefully when the application closes.
    from bluesky_widgets.qt.threading import wait_for_workers_to_quit

    app.aboutToQuit.connect(wait_for_workers_to_quit)

    # Get the catalog (must be databroker.v2-style).
    CATALOG_NAME = "example"
    import databroker

    catalog = databroker.catalog[CATALOG_NAME]
    # Create an instance of our model.
    search_model = SearchAndOpen(catalog, columns=columns)
    # Define what to do when the signal associated with the "Open" button fires.
    # In this toy example, just print to the terminal.
    search_model.events.open.connect(lambda event: print(f"Opening {event.selected_runs}"))
    # Create a Qt "view" of this model...
    search_view = QtSearchListWithButton(search_model)
    # ...and place it in our app.
    central_widget.layout().addWidget(search_view)

    # *** INTEGRATION WITH BLUESKY-WIDGETS ENDS HERE. ***

    # Run the app.
    app.exec_()


if __name__ == "__main__":
    main()



import collections
import logging

from qtpy import QtCore
from qtpy.QtCore import (
    QAbstractTableModel,
    # QItemSelection,
    # QItemSelectionModel,
    QTimer,
    Qt,
)
from qtpy.QtWidgets import QAbstractItemView, QHeaderView, QTableView

from .threading import create_worker


logger = logging.getLogger(__name__)
LOADING_PLACEHOLDER = "..."
CHUNK_SIZE = 5  # max rows to fetch at once
LOADING_LATENCY = 100  # ms


def _load_data(get_data, indexes):
    "Load a batch of data. This is run in a threadpool."
    for index in indexes:
        row, column = index.row(), index.column()
        try:
            item = get_data(row, column)
        except Exception:
            logger.exception("Error while loading search results")
            continue
        yield index, item


class _SearchResultsModel(QAbstractTableModel):
    """
    Qt model connecting our model to Qt's model--view machinery

    This is implementing two layers of "laziness" to ensure that the app
    remains responsive when large tables are loaded.

    1. Rows are added dynamically using Qt's canFetchMore / fetchMore
    machinery.
    2. Data (which Qt assumes is readily available in memory) is immediately
    filled with LOADING_PLACEHOLDER. Work is kicked off on a thread to later
    update this with the actual data.
    """

    def __init__(self, model, *args, **kwargs):
        self.model = model  # our internal model for the components subpackage
        super().__init__(*args, **kwargs)

        # State related to dynamically adding rows
        self._current_num_rows = 0
        self._catalog_length = len(self.model.catalog)

        # Cache for loaded data
        self._data = {}
        # Queue of indexes of data to be loaded
        self._work_queue = collections.deque()
        # Set of active workers
        self._active_workers = set()

        # Start a timer that will periodically load any data queued up to be loaded.
        self._data_loading_timer = QTimer(self)
        # We run this once to initialize it. The _process_work_queue schedules
        # it to be run again when it completes. This is better than a strictly
        # periodic timer because it ensures that requests do not pile up if
        # _process_work_queue takes longer than LOADING_LATENCY to complete.
        self._data_loading_timer.singleShot(LOADING_LATENCY, self._process_work_queue)

        # Changes to the model update the GUI.
        self.model.events.begin_reset.connect(self.on_begin_reset)
        self.model.events.end_reset.connect(self.on_end_reset)

    def _process_work_queue(self):
        if self._work_queue:
            worker = create_worker(_load_data, self.model.get_data, tuple(self._work_queue))
            self._work_queue.clear()
            # Track this worker in case we need to ignore it and cancel due to
            # model reset.
            self._active_workers.add(worker)
            worker.finished.connect(lambda: self._active_workers.discard(worker))
            worker.yielded.connect(self.on_item_loaded)
            worker.start()
        # Else, no work to do.
        # Schedule the next processing.
        self._data_loading_timer.singleShot(LOADING_LATENCY, self._process_work_queue)

    def on_item_loaded(self, payload):
        # Update state and trigger Qt to run data() to update its internal model.
        index, item = payload
        self._data[index] = item
        self.dataChanged.emit(index, index, [])

    def on_begin_reset(self, event):
        # The model is about to set a new catalog with a (potentially)
        # different length and (potentially) different entries. Reset our
        # state.
        self.beginResetModel()
        self.removeRows(0, self._current_num_rows - 1)
        self._current_num_rows = 0
        for worker in self._active_workers:
            # Cease allowing this worker to mutate _data so that we do not get
            # any stale updates.
            worker.yielded.disconnect(self.on_item_loaded)
            # To avoid doing useless work, try to cancel the worker. We do not
            # rely on this request being effective.
            worker.quit()
        self._active_workers.clear()
        self._work_queue.clear()
        self._data.clear()

    def on_end_reset(self, event):
        # The model has its new catalog at this point. Now we can take its
        # length.
        self._catalog_length = len(self.model.catalog)
        self.endResetModel()

    def canFetchMore(self, parent=None):
        if parent.isValid():
            return False
        return self._current_num_rows < self._catalog_length

    def fetchMore(self, parent=None):
        if parent.isValid():
            return
        remainder = self._catalog_length - self._current_num_rows
        rows_to_add = min(remainder, CHUNK_SIZE)
        if rows_to_add <= 0:
            return
        self.beginInsertRows(parent, self._current_num_rows, self._current_num_rows + rows_to_add - 1)
        self._current_num_rows += rows_to_add
        self.endInsertRows()

    def rowCount(self, parent=None):
        return self._current_num_rows

    def columnCount(self, parent=None):
        return len(self.model.headings)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role != Qt.DisplayRole:
            return super().headerData(section, orientation, role)
        if orientation == Qt.Horizontal and section < self.columnCount():
            return str(self.model.headings[section])
        elif orientation == Qt.Vertical and section < self.rowCount():
            return section

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if not index.isValid():  # does > 0 bounds check
            return QtCore.QVariant()
        if index.column() >= self.columnCount() or index.row() >= self.rowCount():
            return QtCore.QVariant()
        if role == QtCore.Qt.DisplayRole:
            if index in self._data:
                return self._data[index]
            else:
                self._data[index] = LOADING_PLACEHOLDER
                self._work_queue.append(index)
                return LOADING_PLACEHOLDER
        else:
            return QtCore.QVariant()


class QtSearchResults(QTableView):
    """
    Table of search results

    Parameters
    ----------
    model: bluesky_widgets.components.search_results.SearchResults
    """

    def __init__(self, model, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model

        self.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.setSortingEnabled(False)
        self.setSelectionBehavior(QTableView.SelectRows)
        self.setShowGrid(True)
        self.setAlternatingRowColors(True)
        self.verticalHeader().setVisible(False)
        # Left alignment prevents column name to move to the right when the last section is stretched
        self.horizontalHeader().setDefaultAlignment(Qt.AlignLeft)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.horizontalHeader().setStretchLastSection(True)
        # self.setAlternatingRowColors(True)
        self._abstract_table_model = _SearchResultsModel(model)
        self.setModel(self._abstract_table_model)

        # Notify model of changes to selection and activation.
        self.selectionModel().selectionChanged.connect(self.on_selection_changed)
        self.clicked.connect(self.on_clicked)

        # Update the view to changes in the model.
        self.model.selected_rows.events.added.connect(self.on_row_added)
        self.model.selected_rows.events.removed.connect(self.on_row_removed)

    def on_selection_changed(self, selected, deselected):
        # One would expect we could ask Qt directly for the rows, as opposed to
        # using set() here, but I cannot find such a method.
        for row in set(index.row() for index in deselected.indexes()):
            if row in self.model.selected_rows:
                self.model.selected_rows.remove(row)
        for row in set(index.row() for index in selected.indexes()):
            if row not in self.model.selected_rows:
                self.model.selected_rows.append(row)

    def on_clicked(self, index):
        self.model.active_row = index.row()

    def on_row_added(self, event):
        """Sync changes to model to view.

        This is expected to be rare, is not yet publicly exposed.
        """
        # TODO -- Not sure what is broken here
        # index1 = self._abstract_table_model.index(event.item, 0)
        # index2 = self._abstract_table_model.index(event.item, self._abstract_table_model.columnCount())
        # selection = QItemSelection(index1, index2)
        # self.selectionModel().select(selection, QItemSelectionModel.Select)
        ...

    def on_row_removed(self, event):
        """Sync changes to model to view.

        This is expected to be rare, is not yet publicly exposed.
        """
        # TODO -- Not sure what is broken here
        # index1 = self._abstract_table_model.index(event.item, 0)
        # index2 = self._abstract_table_model.index(event.item, self._abstract_table_model.columnCount())
        # selection = QItemSelection(index1, index2)
        # self.selectionModel().select(selection, QItemSelectionModel.Deselect)
        ...

    def on_activated_by_model(self, event):
        # TODO
        ...



import importlib.util

import pytest


def pytest_addoption(parser):
    """An option to show windows during tests. (Hidden by default).

    Showing windows decreases test speed by about %18.  Note, due to the
    placement of this conftest.py file, you must specify the bluesky_widgets
    folder (in the pytest command) to use this flag.

    Example
    -------
    $ pytest bluesky_widgets --show-window
    """
    parser.addoption(
        "--show-window",
        action="store_true",
        default=False,
        help="Show window during tests (not shown by default).",
    )


def pytest_collection_modifyitems(session, config, items):
    # When the FigureView fixture returns QtFigure, inject the qtbot fixture
    # as well. This rather invasive hook is needed in order to do this late
    # enough to have access the parameterized tests but early enough to
    # actually ensure that qtbot is applied.

    if importlib.util.find_spec("qtpy"):

        from bluesky_widgets.qt.figures import QtFigure

        for item in items:
            if hasattr(item, "callspec") and "FigureView" in item.callspec.params:
                if item.callspec.params["FigureView"] is QtFigure:
                    item.fixturenames.append("qtbot")
            elif hasattr(item, "callspec") and "FigureViews" in item.callspec.params:
                if item.callspec.params["FigureViews"] is QtFigures:
                    item.fixturenames.append("qtbot")


_figure_view_params = []
if importlib.util.find_spec("qtpy"):

    from bluesky_widgets.qt.figures import QtFigure

    _figure_view_params.append(QtFigure)
if importlib.util.find_spec("ipywidgets"):

    from bluesky_widgets.jupyter.figures import JupyterFigure

    _figure_view_params.append(JupyterFigure)
if importlib.util.find_spec("matplotlib"):

    from bluesky_widgets.headless.figures import HeadlessFigure

    _figure_view_params.append(HeadlessFigure)


@pytest.fixture(params=_figure_view_params)
def FigureView(request):
    return request.param


_figure_views_params = []
if importlib.util.find_spec("qtpy"):

    from bluesky_widgets.qt.figures import QtFigures

    _figure_views_params.append(QtFigures)
if importlib.util.find_spec("ipywidgets"):

    from bluesky_widgets.jupyter.figures import JupyterFigures

    _figure_views_params.append(JupyterFigures)
if importlib.util.find_spec("matplotlib"):

    from bluesky_widgets.headless.figures import HeadlessFigures

    _figure_views_params.append(HeadlessFigures)


@pytest.fixture(params=_figure_views_params)
def FigureViews(request):
    return request.param



from ..search import Search, SearchList


def test_search_list_mutually_exclusive_active_item():
    search_list = SearchList()
    s1 = Search({}, columns=())
    s2 = Search({}, columns=())
    search_list.extend([s1, s2])
    s1.active = False
    s2.active = False
    assert search_list.active is None
    s1.active = True
    assert s1.active
    assert not s2.active
    assert search_list.active is s1
    s2.active = True
    assert not s1.active
    assert s2.active
    assert search_list.active is s2
    s2.active = False
    assert search_list.active is None
    assert "SearchList" in repr(search_list)


def test_search():
    s = Search({}, columns=(), name="test")
    assert s.name == "test"
    assert not s.breadcrumbs
    assert s.root_catalog is s.current_catalog
    assert s.run_search is None
    assert s.results is None
    assert s.input is None
    assert s.active is False
    assert s.active_run is None



"""
Use within IPython like

ipython --gui=qt

In [1]: %run -m bluesky_widgets.examples.ipy_qt_rastered_image
"""
from bluesky_widgets.examples.utils.generate_msgpack_data import get_catalog
from bluesky_widgets.models.plot_builders import RasteredImages
from bluesky_widgets.qt.figures import QtFigure

catalog = get_catalog()
grid_scans = catalog.search({"plan_name": "grid_scan"})
run = grid_scans[-1]

model = RasteredImages("det4", (5, 7))

model.add_run(run)
view = QtFigure(model.figure)
view.show()



"""
In one terminal start a RunEngine worker, which we will send commands to.

start-re-manager --zmq-data-proxy-addr localhost:5577

In another, start a message bus that will forward us Bluesky documents.

bluesky-0MQ-proxy -v 5577 5578

Finally, start this example module, connecting to both of those.

python -m bluesky_widgets.examples.advanced.qt_with_RE_worker tcp://localhost:60615 localhost:5578
"""  # noqa E501

from bluesky_widgets.utils.streaming import stream_documents_into_runs
from bluesky_widgets.qt.zmq_dispatcher import RemoteDispatcher
from bluesky_widgets.models.plot_builders import Lines
from bluesky_widgets.qt.figures import QtFigure
from bluesky_widgets.qt import gui_qt
from bluesky_queueserver.manager.comms import ZMQCommSendThreads
from bluesky_queueserver_api._defaults import default_user_group

import sys
import time


def main():
    with gui_qt("Example App"):
        worker_address, message_bus_address = sys.argv[1:]
        dispatcher = RemoteDispatcher(message_bus_address)
        client = ZMQCommSendThreads(zmq_server_address=worker_address)

        # NOTE: this example starts only if RE Manager is idle and the queue is cleared.
        #   Those are optional steps that ensure that the code in this example is executed correctly.

        # Check if RE Worker environment already exists and RE manager is idle.
        status = client.send_message(method="status")
        if status["manager_state"] != "idle":
            raise RuntimeError(f"RE Manager state must be 'idle': current state: {status['manager_state']}")

        # Clear the queue.
        response = client.send_message(method="queue_clear")
        if not response["success"]:
            raise RuntimeError(f"Failed to clear the plan queue: {response['msg']}")

        # Open the new environment only if it does not exist.
        if not status["worker_environment_exists"]:
            # Initiate opening of RE Worker environment
            response = client.send_message(method="environment_open")
            if not response["success"]:
                raise RuntimeError(f"Failed to open RE Worker environment: {response['msg']}")

            # Wait for the environment to be created.
            t_timeout = 10
            t_stop = time.time() + t_timeout
            while True:
                status2 = client.send_message(method="status")
                if status2["worker_environment_exists"] and status2["manager_state"] == "idle":
                    break
                if time.time() > t_stop:
                    raise RuntimeError("Failed to start RE Worker: timeout occurred")
                time.sleep(0.5)

        # Add plan to queue
        response = client.send_message(
            method="queue_item_add",
            params={
                "item": {
                    "item_type": "plan",
                    "name": "scan",
                    "args": [["det"], "motor", -5, 5, 11],
                },
                "user": "Bluesky Widgets",  # Name of the user submitting item to the queue
                "user_group": default_user_group,
            },
        )
        if not response["success"]:
            raise RuntimeError(f"Failed to add plan to the queue: {response['msg']}")

        model = Lines("motor", ["det"], max_runs=3)
        dispatcher.subscribe(stream_documents_into_runs(model.add_run))
        view = QtFigure(model.figure)
        view.show()
        dispatcher.start()

        response = client.send_message(method="queue_start")
        if not response["success"]:
            raise RuntimeError(f"Failed to start the queue: {response['msg']}")


if __name__ == "__main__":
    main()
